// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import "package:test/test.dart";

import 'package:dotenv/dotenv.dart';

void main() {

  group("Identifier validity", () {
    test("UserId", () {
      assert(Username.isValid("michael-scott"));
      assert(Username.isValid("pit_pattle"));
      assert(Username.isValid("joe"));

      assert(!Username.isValid("#\$^%\$43#\$^"));
      assert(!Username.isValid("wil@ko"));
      assert(!Username.isValid("@matthew"));
    });

    test("Username", () {
      assert(!UserId.isValidFullyQualified("#\$^%\$43#\$^"));
      assert(!UserId.isValidFullyQualified("@pit"));
      assert(!UserId.isValidFullyQualified("pit"));
      assert(!UserId.isValidFullyQualified("pi&*3t"));
      assert(!UserId.isValidFullyQualified("pi&*3:t.tk"));

      assert(UserId.isValidFullyQualified("@pit:pattle.im"));
      assert(UserId.isValidFullyQualified("@joe:matrix.org"));
      assert(UserId.isValidFullyQualified("@jim:192.168.0.1"));
      assert(UserId.isValidFullyQualified("@jim:192.168.0.1:9043"));
    });
  });

  // The intergration tests will use the provided username and password from
  // the .env file, and assume that the given user is a member of the
  // #dart-sdk-test:matrix.org room.
  group("Integration", () {
    Homeserver homeserver;

    LocalUser user;

    setUp(() async {
      homeserver = Homeserver(
        Uri.parse("https://matrix.org")
      );

      // Load .env
      load();
    });

    test('Login', () async {
      user = await homeserver.login(
        Username(env['username']),
        env['password']
      );
    });

    test('Test room exists', () async {

      await for(SyncState syncState in user.sync()) {

        assert(syncState.succeeded);

        final room = await user.rooms[
          RoomId('!otDeInmGHULHzwJyod:matrix.org')
        ] as JoinedRoom;

        assert(room != null);

        final message = await room.timeline.upTo(20)
          .firstWhere((Event event) =>
            event is TextMessageEvent &&
            event.sender.id == UserId('@wilko:matrix.org'));

        assert(message != null);

        final wilko = await room.members[UserId('@wilko:matrix.org')];

        assert(wilko != null);
        assert(wilko.avatarUrl == Uri.parse('mxc://matrix.org/nJHdHeSIgrqBSbvltALNjrEc'));

        final shipment = room.send(TextMessage(body: 'Sent from the Dart Matrix SDK'));
        await for (final sentState in shipment) {
          final sentEvent = await room.timeline.upTo(1).first as TextMessageEvent;
          assert(sentEvent.content.body == 'Sent from the Dart Matrix SDK');
        }

        final avatarBytes = await homeserver.download(room.avatarUrl);
        assert(avatarBytes.length == 95245);

        final thumbnailBytes = await homeserver.downloadThumbnail(
          room.avatarUrl,
          width: 64,
          height: 64
        );

        final imageMessage = await room.timeline.upTo(200)
            .firstWhere((event) => event is ImageMessageEvent) as ImageMessageEvent;

        final imageBytes = await homeserver.download(imageMessage.content.url);
        assert(imageBytes.length == 155471);

        assert(imageMessage.content.info.width == 720);
        assert(imageMessage.content.info.height == 1280);

        break;
      }

      await user.logout();
      assert(user.isLoggedOut);
    }, timeout: const Timeout(const Duration(minutes: 5)));
  });
}
